#include <forms.h>
#include "synth.h"
#include <stdio.h>
#include <math.h>
#include "st.h"
#include "libst.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
/**************************************************************************
        Prototypes
************************************************************************/
void display() , cleanup();
void scale(FL_OBJECT *,long);
void ossdspstartread();  /* ossdsp functions*/
LONG  ossdspread();
void ossdspstopread();
void ossdspstartwrite();
void ossdspwrite();
void ossdspstopwrite();
void dspreset();
void resetposn();

/**************************************************************************
        Constants
************************************************************************/
#define PI 3.1415926
#define NMAX 1024           /* Number of data points in plots*/
#define NPLAY 2048             /*Number of data points in play array*/

#define RATE 22050        /*Number of data per second to sound card*/
#define PLAYRATE 44100    /* Rate at which to play sounds through sound card*/
#define SECONDS_R 2.0  /*Number of seconds of recording*/
#define SECONDS_W 2.0  /* Seconds of sound */
#define NCYC 10       /* Numbr of times recording is updated per second*/

/**************************************************************************
       Globals
************************************************************************/

FD_synth * ui;      /*frame name*/
Window synth_window; /*Name of main window.*/
double ampl[2][32],mampl[3][2][32]; /* harmonic amplitues and phases, and
                                       memorized data*/
float four[2][NMAX],dum[NMAX];  /* Fourier components and imaginary dummy data
                                      for fft */
float y[NMAX],x[NMAX],Triang[2][32],Square[2][32],Vowel[2][32],pampl;
FL_OBJECT *Amp[32];           /* Amplifude slider pointers*/
FL_OBJECT  *P[32];           /* Phase slider pointers*/
FL_OBJECT  *M[4], *Z[2];     /* Memory button and zoom button pointers*/
int memflag[4]={0,0,0,0};    /* m[0]= memory button, m[1-3]= memory present*/
int recordstate;             /* state of recorder. 0= not opertional, 1= ready
                             (or display old recoding) 2= recording*/
int Nfr=2;        /* Number of cycles of data displayed */
int fourier_flag=0;       /*To display or to hide spectrum*/
int ll_flag=FL_LINEAR;           /*Linear or Log scale for spectrum*/
double rectime;             /*Length of recording time (.1 to 2 sec)*/
int level=0, maxlevel=-1,xvals=0; /* recording level, max recorded level,*/
double freqmag;                 /* amplification for amplitude sliders */
double average_read;           /* average of the read input. */

LONG yr[(long)SECONDS_R*RATE];           /* oss routine data buffer*/
float xrd[(long)SECONDS_R*RATE],yrd[(long)SECONDS_R*RATE];  /* normalised data for display*/
float xmax[5],xmin[5];  /*min and max values of selected areas on recorded data*/
int inputflag;
char * Filename;
int loadflag=0,saveflag=0;
int grab_ok_flag=0;  /* values 0,1,2-- 0=test, 1=OK, 2=cancel*/

char *ossdspnames[] = { "ossdsp", (char *) 0 };

format_t ossformat={ossdspnames, 0 /* Mono*/,
                &ossdspstartread, &ossdspread, &ossdspstopread,     /* /dev/dsp */
                &ossdspstartwrite, &ossdspwrite, &ossdspstopwrite};
struct soundstream ss;

/**************************************************************************
       Procedure main
************************************************************************/

main(int argc,char * argv[])
{
  
   int i;
   /* Initialize form library, create and display main form ... */
   fl_initialize(&argc, argv, "synth", 0, 0);
   ui = create_form_synth();
   fl_hide_object(ui->inputgroup);
   fl_hide_object(ui->fourier);
   fl_set_input_return(ui->filename,FL_RETURN_END);
   fl_hide_object(ui->invalidgp);
   fl_hide_object(ui->grab_warn_gp);
    synth_window=fl_show_form(ui->synth,
                FL_PLACE_CENTER | FL_FREE_SIZE, 
                FL_FULLBORDER,
                "synth");
  
		
   /* Make aray for the amplitude Phase and memory OBJECTS */		
   Amp[1]=ui->amp1;
   Amp[2]=ui->amp2;
   Amp[3]=ui->amp3;
   Amp[4]=ui->amp4;
   Amp[5]=ui->amp5;
   Amp[6]=ui->amp6;
   Amp[7]=ui->amp7;
   Amp[8]=ui->amp8;
   Amp[9]=ui->amp9;
   Amp[10]=ui->amp10;
   Amp[11]=ui->amp11;
   Amp[12]=ui->amp12;
   Amp[13]=ui->amp13;
   Amp[14]=ui->amp14;
   Amp[15]=ui->amp15;
   Amp[16]=ui->amp16;
   Amp[17]=ui->amp17;
   Amp[18]=ui->amp18;
   Amp[19]=ui->amp19;
   Amp[20]=ui->amp20;
   Amp[21]=ui->amp21;
   Amp[22]=ui->amp22;
   Amp[23]=ui->amp23;
   Amp[24]=ui->amp24;
   Amp[25]=ui->amp25;
   Amp[26]=ui->amp26;
   Amp[27]=ui->amp27;
   Amp[28]=ui->amp28;
   Amp[29]=ui->amp29;
   Amp[30]=ui->amp30;
   P[1]=ui->p1;
   P[2]=ui->p2;
   P[3]=ui->p3;
   P[4]=ui->p4;
   P[5]=ui->p5;
   P[6]=ui->p6;
   P[7]=ui->p7;
   P[8]=ui->p8;
   P[9]=ui->p9;
   P[10]=ui->p10;
   P[11]=ui->p11;
   P[12]=ui->p12;
   P[13]=ui->p13;
   P[14]=ui->p14;
   P[15]=ui->p15;
   P[16]=ui->p16;
   P[17]=ui->p17;
   P[18]=ui->p18;
   P[19]=ui->p19;
   P[20]=ui->p20;
   P[21]=ui->p21;
   P[22]=ui->p22;
   P[23]=ui->p23;
   P[24]=ui->p24;
   P[25]=ui->p25;
   P[26]=ui->p26;
   P[27]=ui->p27;
   P[28]=ui->p28;
   P[29]=ui->p29;
   P[30]=ui->p30;

   M[1]=ui->mem1;
   M[2]=ui->mem2;
   M[3]=ui->mem3;
   
   Z[0]=ui->zoom1;
   Z[1]=ui->zoom2;




   



  /* initialise variables */
   for (i=0;i<NMAX;i++)
      { 
        four[0][i]=four[1][i]=0;
	x[i]=i;
       }
   for(i=0;i<(long)SECONDS_R*RATE;i++) xrd[i]=i;
   for (i=0;i<32;i++)
      {
        ampl[0][i]=ampl[1][i]=0;
       }
  /* initialising graph */
   fl_set_xyplot_xtics(ui->xyplotgr,-1,-1);
   scale(ui->scale,0);
   fl_set_positioner_ystep(ui->positioner,100.);
  

  
  /* Initialise the ossdsp soundstream structure*/  
 
  ss.filename="/dev/dsp";
  ss.filetype="ossdsp";
  ss.swap=0;
  ss.h   =&ossformat;
  ss.info.size=WORD;
  ss.info.rate=22050;
  ss.info.style=SIGN2;
  
   
   /* Enter main forms loop ... */             
   fl_do_forms();
   fl_finish();
   
   /* close /dev/dsp */
   cleanup();
   
   return 0;
}

/*************************************************************************
     Function cleanup-- close the /dev/dsp file
*************************************************************************/

void cleanup()
{
        /* Close the input file and outputfile before exiting*/
        if (ss.fp)
           {
	        fclose(ss.fp);
		ss.fp=(FILE *)NULL;
	   }
}


/************************************************************************
       Function amp  -- Amplitude Slider changed
************************************************************************/

void amp(FL_OBJECT * aslide, long val)
{
   ampl[0][val]= fl_get_slider_value(aslide);
   display();   
}

/**************************************************************************
       Function phase  -- Phase Slider changed
************************************************************************/
/* Phase slider (0-1) changed */
void phase(FL_OBJECT * pslide, long val)
{
   ampl[1][val]= fl_get_slider_value(pslide);
   display();
}
/**************************************************************************
       Function play  -- Play note on button push
************************************************************************/
/* Uses the ossdsp routines from sox*/
void play(FL_OBJECT * playbut, long val)
{
     int i,j;
     float rate,Cmax;
     long Pmax;
     static double dt,t=0;
     double playtime;
     float amax,vol,maxnumval;
     int NPCYCL=16; /* Number of periods per full cycle)*/
     FILE * fd;
     
     double frequency,tmp;
     
     frequency=fl_get_slider_value(ui->freq);
     
     tmp=(double)( frequency*NMAX/((double)PLAYRATE));
     
   /*  if(tmp<=2.) NPCYCL=2;
     else
     {
        if(tmp<4.)NPCYCL=4;
	else{if(tmp<=8.) NPCYCL=8;
	      else NPCYCL=16;
	    }
     }
    */
    NPCYCL=1;
     rate=(frequency*(double)4*NMAX/(double)NPCYCL);
     
 //    if(rate<PLAYRATE/2) rate=PLAYRATE/2;
//     if(rate>PLAYRATE) rate=PLAYRATE;
    /* ss.info.rate=rate;*/
       ss.info.rate=PLAYRATE;

     
     dt=1./rate;
     for(i=0;i<NMAX;i++)
        {
         four[0][i]=0;
         four[1][i]=0;
         x[i]=(float)i;
        }
     maxnumval=fl_get_slider_value(ui->maxnumslide);
     for (i=1;i<=maxnumval;i++)
        {
         
          {      
            four[0][NMAX-NPCYCL*i]=four[0][NPCYCL*i]=ampl[0][i]  *cos(ampl[1][i]*PI*2)/2 ;
            four[1][NPCYCL*i]=ampl[0][i]*sin(ampl[1][i]*PI*2)/2;
            four[1][NMAX-NPCYCL*i]=-ampl[0][i]*sin(ampl[1][i]*PI*2)/2;
          }
        }
     fft(four[0],four[1],y,dum,NMAX,-1); 
     amax=1.;
/*    Find the maximum value of y-- do not know why I put this in here.*/
/*
     amax=0;
     for (i=0;i<NMAX;i++)  
       {
         if(amax<fabs(y[i])) amax=fabs(y[i]);
       }
*/
     playtime=fl_get_slider_value(ui->playtime); /*How long to play (max 2sec)*/
       
     vol=fl_get_slider_value(ui->volume); /*Volume of tone*/
     if( fl_get_button(ui->ampbut)==1) /* also amplify by display scale ? */
        pampl=exp(fl_get_slider_value(ui->scale)*log(2));
     else 
        pampl=1.;
       
       /* Open /dev/dsp and get ready to write */
       
     if ((fd=fopen("/dev/dsp","r+"))==(FILE *)NULL)
       { 
	  fprintf(stderr,"Cannot open /dev/dsp\n");
	  return;
       } 

    {
         float zi,amplitude=vol/(2.*amax*pampl)*pow(2.,(double)(8*sizeof(LONG)-1));
         
	 Cmax=floor( 4.*(double)NMAX*frequency/PLAYRATE);
	 Pmax=Cmax*PLAYRATE/frequency;
	 for (i=0;i<Pmax;i++)
	   {
	       long j;
	       zi=((double)i/(double)Pmax)*Cmax;
	       zi=(zi-floor(zi))*NMAX;
	       j=zi;
	       yr[i]=amplitude*( y[j]*(-zi+j+1)
	                +y[(j+1)%NMAX]*(zi-j));
	   }	   
     }
     ss.fp=fd;
     ossdspstartwrite(&ss);
/*       
     for(i=0;i<NMAX;i++)
          {
	    yr[i]=(LONG)((double)(vol*y[i]/(2.*amax*pampl)
	           *pow(2.,(double)(8*sizeof(LONG)-1))));
	   }
*/	   
     /* Testing output	   
	for(i=0;i<10;i++)printf("yr=%ld, y=%g,pow= %g, amax=%g,pampl= %g,vol=%g,yrd=%g\n"
	,yr[i],(double)y[i],
	  pow(2.,(double)(8*sizeof(LONG)-1.)),(double)amax,(double)pampl,(double)vol,
          (double)(vol*y[i]/(2.*amax*pampl) *pow(2.,(double)(8*sizeof(LONG)-1))));
     */

     for(j=0;j<(int)(playtime*frequency/Cmax);j++)
         {
	   ossdspwrite(&ss,yr,(LONG)Pmax);
	 }
     ossdspstopwrite(&ss);
     cleanup();
	   
     /*
     for (j=0;j<(int)((playtime*RATE)/NMAX);j++)  
     
      {
      
       for (i=0;i<NMAX;i++)
         printf( "%g  %g\n", (t+=dt),vol*y[i]/(2.*amax*pampl));
      }
     for(j=0;j<10;j++)
       for(i=0;i<NMAX;i++)
          printf("%g  %g\n",(t+=dt),0.);
     */      
}

/**************************************************************************
       Function dummy_callback  -- dummy null routine
************************************************************************/
/* dummy callback routine. Sliders without callback crash */
void dummy_callback(FL_OBJECT * vv, long val) {}

/**************************************************************************
       Function vowel  -- Predefined oboe sound (oboe button)
************************************************************************/
/*Set up oboe or vowel sound ( two formants, 
one at about 2 freq and one at 11*/
void vowel(FL_OBJECT * vbut, long val)
{
   int i;
   for (i=1;i<31;i++)
     { ampl[0][i]=0;
       ampl[1][i]=0;
     }
   ampl[0][1]=.1;
   ampl[0][2]=.8;
   ampl[0][3]=.6;
   ampl[0][4]=.2;
   ampl[0][11]=.04;
   ampl[0][12]=.05;
   ampl[0][13]=.02;
   ampl[0][14]=.01;
   for(i=1;i<31;i++)
     {
       fl_set_slider_value(Amp[i],ampl[0][i]);
       fl_set_slider_value(P[i],ampl[1][i]);      
     }
   display();
}

/**************************************************************************
       Function reset- Set all Slider to 0
************************************************************************/
/*Set all amplitudes to 0 and all phases to 0*/
void reset(FL_OBJECT * resetbut, long val)
{
   int i;
  
   for (i=1;i<31;i++)
     {
      ampl[0][i]=ampl[1][i]=0;
      fl_set_slider_value(Amp[i],ampl[0][i]);
      fl_set_slider_value(P[i],ampl[1][i]);
      }
   display(); 
      
}

/**************************************************************************
       Function scale -- slider to rescale display
************************************************************************/
/* Rescale the output */
void scale(FL_OBJECT * scaleslide, long val)
{
        pampl=exp(fl_get_slider_value(scaleslide)*log(2));
	fl_set_xyplot_ybounds(ui->xyplotgr,-pampl,pampl);
	fl_set_xyplot_xbounds(ui->xyplotgr,0.,1024.);
	clip();
}

/**************************************************************************
       Function triang -- predefined triangle wave
************************************************************************/
/* Triangle wave */
void triang(FL_OBJECT * triangbut, long val)
{
   int i;
   for (i=1;i<31;i++)
    { ampl[0][i]=(i%2)/(double)(i*i);
      ampl[1][i]=0;
      fl_set_slider_value(Amp[i],(double)ampl[0][i]);
      fl_set_slider_value(P[i],ampl[1][i]);
    }
   display();
}

/**************************************************************************
       Function square-- predefined square wave with duty cycle
************************************************************************/
/*square wave with duty cycle */
void square(FL_OBJECT * squarebut, long val)
{
  int i;
  double duty;
  duty=fl_get_slider_value(ui->sqduty);
  for (i=1;i<31;i++)
    { ampl[0][i]=Square[0][i]=fabs(sin(PI*i*duty)/i);
      Square[1][i]=0;
      if (sin(PI*i*duty)<0) (Square[1][i]+=.5);
      ampl[1][i]=(Square[1][i]-=(int)Square[1][i]);
      fl_set_slider_value(Amp[i],ampl[0][i]);
      fl_set_slider_value(P[i],ampl[1][i]);
    }
  display();
 
}

/**************************************************************************
       Function sqslide-- change duty cycle of square wave
************************************************************************/
/* slider for duty cycle of square wave*/
void sqslide(FL_OBJECT *sqduty, long val)
{
   square(ui->square,0);
}

/**************************************************************************
       Function sawbut -- Predefined sawtooth wave
************************************************************************/
/* Saw Tooth wave */ 
void sawbut(FL_OBJECT *saw,long val) 
{
   int i;
   for (i=1;i<31;i++)
     { ampl[0][i]=1./i;
       ampl[1][i]=.25;
       fl_set_slider_value(Amp[i],ampl[0][i]);
       fl_set_slider_value(P[i],ampl[1][i]);      
     }
   display();
  
}
 
/**************************************************************************
       Function memory -- memory buttons callback
************************************************************************/
/*Memorize current display, or display remembered spectrum*/ 
void memory( FL_OBJECT *mem,long val) 
 {
     int i;
     /* Memory button pressed if val=0*/
     
     if(val==0) 
       memflag[0]^=1;
     else
      {
        if(memflag[0]==0)      /* Memory button not pressed. Display*/
	 {                     /* remembered spectrum */
	 if(memflag[val]==1)
          {
	   for(i=1;i<31;i++)
	    {
	     ampl[0][i]=mampl[val-1][0][i];
             ampl[1][i]=mampl[val-1][1][i];
	     fl_set_slider_value(Amp[i],ampl[0][i]);
	     fl_set_slider_value(P[i],ampl[1][i]);
	    }
	   display();
	  }
	 }
	else                 /*Memory button pressed-- store current spectrum*/
	 {
	   memflag[val]=1;   /* flag that spectrum in memory*/
	   memflag[0]=0;     /*cancel memory button light*/
	   fl_set_object_boxtype(M[val],FL_UP_BOX);
	   for(i=1;i<31;i++)
	     {
	       mampl[val-1][0][i]=ampl[0][i];
	       mampl[val-1][1][i]=ampl[1][i];
	       
	     }
	  }
     
       }
   
      
     fl_set_button(ui->mem,memflag[0]);  /* Set light as to whether to store*/
	                                    /* or display*/
}

/**********************************************************************
       Function cycle_button
       Change number of cycles displayed, or number of cycles
              in grabbeddata
***********************************************************************/
void cycle_button(FL_OBJECT * cyclebut, long val)
{
    if(recordstate==0)
      display();

}
/**************************************************************************
       Function maxnum -- slider for max harmonic displayed
************************************************************************/ 
 /* Set the display scaling slider*/
void maxnum(FL_OBJECT * maxnumslide, long val)
   {
     display();
   }
/************************************************************************
  Function display() --Do Fourier transform and display graph
*************************************************************************/   
void display()
 {
   int i;
   
   /* return ampl sliders to normal position */
   if(freqmag>1.)
      {
        freqmag=1.;
        fl_set_slider_value(ui->freqsize,freqmag);
        for (i=1;i<31;i++)
          fl_set_slider_value(Amp[i],ampl[0][i]); 
      }
      
    for(i=0;i<NMAX;i++)
      {
       four[0][i]=0;
       four[1][i]=0;
       x[i]=(float)i;
      }
    Nfr=fl_get_counter_value(ui->cyclebut);
    for (i=1;i<31;i++)
      {
        if ( i <= fl_get_slider_value(ui->maxnumslide))
         {      
           four[0][NMAX-Nfr*i]=four[0][Nfr*i]=ampl[0][i]  *cos(ampl[1][i]*PI*2)/2 ;
           four[1][Nfr*i]=ampl[0][i]*sin(ampl[1][i]*PI*2)/2;
           four[1][NMAX-Nfr*i]=-ampl[0][i]*sin(ampl[1][i]*PI*2)/2;
         }
      }
    fft(four[0],four[1],y,dum,NMAX,-1);
    fl_set_xyplot_data(ui->xyplotgr,x,y,NMAX,"","","");
    fl_delete_xyplot_overlay(ui->xyplotgr,1);

    clip();
} 
 
 
/**************************************************************************
       Function xygraph -- callback to changes in active xygraph (deprecated)
************************************************************************/ 
void xygraph(FL_OBJECT * recgraph, long val)
 { 
   int i;
   float x,y,xloc;
   FL_Coord sx,sy;
   static xnumb=0;
   if(recordstate==0)
     {fl_delete_xyplot_overlay(ui->xyplotgr,1);

      return;
     }
   /*   printf("xygraph start level=%d maxlevel=%d \n", level, maxlevel);*/
   if (level==4) return;
   (void)fl_get_win_mouse(FL_ObjWin(ui->xyplotgr),&sx,&sy,&i);
   fl_xyplot_s2w(ui->xyplotgr,sx,sy,&x,&y);
   fprintf(stderr,"sx=%g sy=%g x=%g y=%g\n",sx,sy,x,y);
   /*fl_get_xyplot(ui->xyplotgr, &x, &y, &i);*/

   xloc=xmin[level]+(x/(double)NMAX)*(xmax[level]-xmin[level]);
   
   /*   fprintf(stderr,"x=%g y=%g i=%d xmin=%g xmax=%g,xloc=%g\n"
        ,x,y,i,xmin[level],xmax[level],xloc);
   */
   if (xmin[level+1]<0) 
     {
       xmin[level+1]=xloc;       
       xmax[level+1]=-1;}
   else 
     {
       if (xmax[level+1]<=0.) 
         {
	   xmax[level+1]=xloc;	   
	 }
       else 
         {
	   xmin[level+1]=xmax[level+1]; 
	   xmax[level+1]=xloc;
	 }
     }
   showgr();
  
   if(level<3)
     {
       for (i=level+2;i<5;i++)
        {
          xmin[i]=xmax[i]=-1;
        }
       maxlevel=level;
     }
}

/**************************************************************************
       Function position-- positioner marking of spectrum
************************************************************************/
void position(FL_OBJECT * positioner,long val)
{
  
   int i;
   float x,y,xloc,a,b;
   FL_Coord cx,cy,w,h;
   static xnumb=0;
   if(recordstate==0)
     {fl_hide_object(ui->positioner);
      return;
     }
   /*   printf("xygraph start level=%d maxlevel=%d \n", level, maxlevel);*/
   if (level==4)
      {
       resetposn();
       return;
       }
//   x=fl_get_positioner_xvalue(ui->positioner);
   
//   fl_get_xyplot_xmapping(ui->xyplotgr,&a,&b);
   /*fl_get_object_geometry(ui->xyplotgr,&cx,&cy,&w,&h); */
//   fl_get_object_geometry(ui->positioner,&cx,&cy,&w,&h);
   /*fprintf(stderr,"a=%g b=%g\n",a,b);*/
//      xloc=xmin[level]+(x*w-b+cx)/(1.006*a*(NMAX))*(xmax[level]-xmin[level]);  
         /* 600= length of xyplot screen*/
         /* The 1.006 is a kludge-- there seems to be some problem
	     with the screen mapping values, and they give the wrong answer.
	     (borders??)
	 */

   {
      FL_Coord px,py;
      float X,Y;
      unsigned int k;
      fl_get_win_mouse(synth_window,&px,&py,&k);
      fl_xyplot_s2w(ui->xyplotgr, (double )px, (double) py,&X,&Y);
      xloc=xmin[level]+X/(NMAX+1.) *(xmax[level]-xmin[level]);
 
   }  


   if(xloc<xmin[level]) xloc=xmin[level];
   if(xloc>xmax[level]) xloc=xmax[level];

   if (xmin[level+1]<0) 
     {
       xmin[level+1]=xloc;       
       xmax[level+1]=-1;}
   else 
     {
       if (xmax[level+1]<=0.) 
         {
	   xmax[level+1]=xloc;	   
	 }
       else 
         {
	   xmin[level+1]=xmax[level+1]; 
	   xmax[level+1]=xloc;
	 }
     }
   showgr();
   resetposn();
   if(level<3)
     {
       for (i=level+2;i<5;i++)
        {
          xmin[i]=xmax[i]=-1;
        }
       maxlevel=level;
     }
}

/**************************************************************************
       Function resetposn -- reset positioner off screen.
************************************************************************/
void resetposn()
{
   fl_hide_object(ui->positioner);
   fl_set_positioner_yvalue(ui->positioner,100.);
   fl_set_positioner_xvalue(ui->positioner,0.);
   fl_show_object(ui->positioner);
}

/**************************************************************************
       Function zoom -- in and out buttons to zoom in on spectrum
************************************************************************/
void  zoom(FL_OBJECT * zoombut, long val)
 {
   if(recordstate==0)return;
   if(val==1)
    {

       if ( level>3||maxlevel<0)
          return;
       if(xmax[level+1]<0)
          return;
       level++;
       if(level>maxlevel)
          maxlevel=level;
       if(xmax[level]<xmin[level])
         {
           float dum;
	   dum=xmin[level];
	   xmin[level]=xmax[level];
	   xmax[level]=dum;
         }

     }
    else
     {
        if(level==0) 
	   return;
        level--;
      
     }
     showgr();
     resetposn();
     fourier_flag=(fourier_flag+1)%2;
     fourier_but(ui->fourier_button,0);
     fl_set_positioner_xvalue(ui->spectrum_posn,0.);  
}
 
/**************************************************************************
       Function grab-- get current recorded display into spectrum
************************************************************************/
void  grab(FL_OBJECT * grabbut,long val)
{
   double dx,xlim;
   int cyc,i;
   
   if(recordstate==0||grab_ok_flag==2)
     {
       grab_ok_flag=0;
       return;
     }
    if(grab_ok_flag==0)
     {
       if (level==0||(level<4 && xmin[level+1]>=0&&xmax[level+1]>=0))
        {  
	  fl_show_object(ui->grab_warn_gp);
          return;
	 }
     }
   
   /* Reset recordstate and record button*/
   recordstate=0;
   grab_ok_flag=0;
   fl_set_object_color(ui->recordbut,FL_COL1,FL_LIGHTER_COL1);
   fl_hide_object(ui->positioner);
   fl_delete_xyplot_overlay(ui->xyplotgr,1);

   fourier_but(ui->fourier_button,0);  
   /* Check on the level-- do not use on lowest level*/
   
   if(maxlevel<0||maxlevel>4||level==0) return;
   
   if(xmax[level]<xmin[level])      /* order the xmin and xmax*/
     { 
       float dum;
       dum=xmax[level];
       xmax[level]=xmin[level];
       xmin[level]=dum;
     }
   dx=(xmax[level]-xmin[level])/(double)NMAX;  /*delta screen point dx is 1/RATE
                                                   ticks*/
   xlim=xmin[level];
   
   /*get the amplitude slider from left of screen if button is on*/
   if( fl_get_button(ui->ampbut)==1)
        pampl=exp(fl_get_slider_value(ui->scale)*log(2));
   else 
        pampl=1.;
   
   for (i=0;i<NMAX;i++)
     {
        y[i]=((1-(xlim-(double)((int)xlim)))*yrd[ (int)(xlim)] 
             +( xlim-(double)((int)xlim))*yrd[(int)xlim +1])/pampl;
        xlim+=dx;
        dum[i]=0.;
     }
   fft(y,dum,four[0],four[1],NMAX,-1);
   cyc=fl_get_counter_value(ui->cyclebut)+.001;
   for(i=1;i<31;i++)
     {
       ampl[0][i]=2.*sqrt((double)(four[0][cyc*i]*four[0][cyc*i]
                 +four[1][cyc*i]*four[1][cyc*i]))/NMAX;
       if( ampl[0][i]> .00001)
         {
	   ampl[1][i]=acos((double)(2.*four[0][cyc*i]/(NMAX*ampl[0][i])))/(2*PI);
	   if(four[1][cyc*i]>0)
	      ampl[1][i]=1.-ampl[1][i];
	 }  
       else
         ampl[1][i]=0.;
       fl_set_slider_value(Amp[i],ampl[0][i]);
       fl_set_slider_value(P[i],ampl[1][i]);

     } 
     /*set slider frequency*/
     {
       float frequency;
       frequency=cyc*(double)RATE/(double)(xmax[level]-xmin[level]);

       fl_set_slider_value(ui->freq,frequency);
       fourier_but(ui->fourier_button,0);
     }
}

/**************************************************************************
       Function recorder -- record from /dev/dsp
************************************************************************/ 
void recorder(FL_OBJECT * recordbut, long val)
{
   FILE * fd;
   recordstate= (recordstate +1)%3;

   if(recordstate==1)
     {
        fl_set_object_color(ui->recordbut,FL_GREEN,FL_LIGHTER_COL1);
        if(maxlevel<0)
	  return;
        level=0;
	showgr();
	resetposn();
	return;
     }
	
   if(recordstate==2)
      {
        long nread,i;
        fl_set_object_color(ui->recordbut,FL_YELLOW,FL_LIGHTER_COL1);
	/* Stick in the next line so as to give the button time to change
	   colour. Otherwise it sits on its old colour until the dsp has been
	   read
	*/ 
	fl_redraw_object(ui->recordbut);
       
        maxlevel=0;
        level=0;
        for (i=0;i<5;i++)
           xmax[i]=xmin[i]=-1.;
	 
        if ((fd=fopen("/dev/dsp","r+"))==(FILE *)NULL)
          { 
	    fprintf(stderr,"Cannot open /dev/dsp\n");
	    recordstate=0;
	    fl_set_object_color(ui->recordbut,FL_COL1,FL_LIGHTER_COL1);
            fl_hide_object(ui->positioner);
            display();
            return;
          } 
        ss.fp=fd;
        ss.info.rate=RATE;
        ossdspstartread(&ss);
        bzero(yr,sizeof(long)*rectime*RATE);
        bzero(yrd,sizeof(float)*rectime*RATE);
        xmin[0]=0;
        rectime=fl_get_slider_value(ui->rectime);
        xmax[0]=rectime*RATE-1;
        {
          long *yrp,j,jrp=0;
	  double total_read=0.;
	  yrp=yr;
          for (i=0;i<(int)(NCYC*rectime);i++)
           { 
	     nread=ossdspread(&ss,yrp,(int)RATE/NCYC);
	     if(nread<RATE/NCYC)
	         fprintf(stderr," incomplete read %ld %ld\n",i,nread);
	     for (j=jrp;j<jrp+RATE/NCYC;j++)
	        {
		   yrd[j]=(float)yr[j]/pow((double)2.,(double)31.);
		   total_read+=yrd[j];
		}
             yrp= yrp +(RATE/NCYC);
	     jrp+=RATE/NCYC;
	     average_read=total_read/(double)jrp;
	   
	   
	     showgr();
	   }
        
        }
        cleanup();
       
        fl_set_object_color(ui->recordbut,FL_RED,FL_LIGHTER_COL1);
       
        /* for(i=0;i<nread;i++)
          {
	    yrd[i]=(float)yr[i]/pow((double)2.,(double)31.);
	  }
        if(nread<(long)(SECONDS_R*RATE))
          for(i=nread;i<(long)(SECONDS_R*RATE);i++)
	     yrd[i]=0.;
        */
        resetposn();
        xmin[0]=0.;
        xmax[0]=rectime*RATE-1.;
        /*       showgr();
        */
        return;
     }
       
   
   /* recordstate=0 --ie the system drops out of the record state.*/ 
   {
   /*printf("recordstate= %d\n",recordstate);*/
     fl_set_object_color(ui->recordbut,FL_COL1,FL_LIGHTER_COL1);
     fl_hide_object(ui->positioner);
     fourier_but(ui->fourier_button,0);
     display();
     return;
   }
}


 
/**************************************************************************
       Function showgr -- display the recorded sound.
************************************************************************/ 
 
 
showgr()
{
    int i;
    char xlabel[100];
    double dx,xlim;
    float xp[NMAX];
    if(xmax[level]==xmin[level])
         xmin[level]=xmax[level]-1./(double)NMAX;
    dx=(xmax[level]-xmin[level])/(double)(NMAX);
    xlim=xmin[level];
    if(dx<0)
      {
        dx=-dx;
	xlim=xmax[level];
      }
    if(xlim<1) xlim=1;
    

    sprintf(xlabel,"Xmin=%g   Xmax=%g   Fx=%g\0"      
          ,(double)(xmin[level]/(RATE)),(double)(xmax[level]/(RATE))
          ,(double)(RATE/(xmax[level]-xmin[level])));
	     
   if (dx>2.)
     { 
       int i;
       long j,k;
       double ytmpmin,ytmpmax;
      
       dx=dx*(double)(NMAX)/(double)(NMAX+2.);

       for (i=0;i<NMAX/2;i++)
         {
	  
           if( xlim-(k=(long)xlim)>0)
              k++;
           ytmpmin=ytmpmax=yrd[k]; 
	   
           for(j=k+1;j<(xlim+2.*dx);j++)
             {
               
                 if(ytmpmax<yrd[j]) ytmpmax=yrd[j];
                 if(ytmpmin>yrd[j]) ytmpmin=yrd[j];
             } 

           y[2*i]=ytmpmax-average_read;
           y[2*i+1]=ytmpmin-average_read;
 	   xp[2*i]=2*i;
	   xp[2*i+1]=2*i;
	      
           xlim+=2.*dx;
         }

	fl_set_xyplot_data(ui->xyplotgr,xp,y,NMAX,"",xlabel,"");
	
     }
   else
     {
       for (i=0;i<NMAX;i++)
         { 
            y[i]=(1.- (xlim-(long)xlim))*yrd[(long)xlim] +
                   (xlim-(long)xlim)*(yrd[(long)xlim+1])-average_read;
	    xp[i]=i;
            xlim+=dx; 
          }

	fl_set_xyplot_data(ui->xyplotgr,xp,y,NMAX,"",xlabel,"");
     }
  

   if((maxlevel>level)||((level<4) && (xmin[level+1]>=0)))
      { 
        int npts;
        float xo[2],yo[2]={8.,8.}; 
        xo[0]=(float)NMAX-(xlim-xmin[level+1])/dx;
        npts=1;
        if(xmax[level+1]>=0)
          {
            xo[1]=(float)NMAX-(xlim-xmax[level+1])/dx;
            npts=2;
	    /*fprintf(stderr,"xo0=%g xo1=%g\n",xo[0],xo[1]);*/
          }
	    
        /*fl_set_xyplot_overlay_type(ui->xyplotgr,1,FL_LINEPOINTS_XYPLOT);*/
        fl_set_xyplot_overlay_type(ui->xyplotgr,1,FL_IMPULSE_XYPLOT);
        fl_add_xyplot_overlay(ui->xyplotgr,1,xo,yo,npts,FL_RED);
      }
   else
      {
        fl_delete_xyplot_overlay(ui->xyplotgr,1);
      }  
    
}



/**************************************************************************
       Function freqvol -- amplify the displayed amp slider positions.
************************************************************************/
void freqvol(FL_OBJECT * vol, long val)
{
    int i;
    freqmag=   fl_get_slider_value(ui->freqsize);
    for (i=1;i<31;i++)
      fl_set_slider_value(Amp[i],freqmag*ampl[0][i]);
}

/**************************************************************************
       Function ampapply - dummy function for display amplitude slider affecting
                            grabbing and playing
************************************************************************/
void ampapply(FL_OBJECT * ampbut,long val)
{
   clip();
}

/**************************************************************************
       Function filename -- get filename for save or load file.
************************************************************************/
void filename(FL_OBJECT *inputfilename, long val)
{
    inputflag=1;
    if(loadflag==1) 
      {
        load(ui->loadbut,0);
        return;
      }
    if(saveflag==1)
      {
        save(ui->savebut,0);
	return;
      }
    loadflag=saveflag=0;
    fl_hide_object(ui->inputgroup);
    fl_hide_object(ui->invalidgp);
}

/**************************************************************************
       Function load -- read spectrum from file.
************************************************************************/
void load (FL_OBJECT * loadbut, long val)
{
    FILE * fd;
    int i;
    char dummy[85];
    if(loadflag==0)
      {     
        fl_show_object(ui->inputgroup);
        loadflag=1;
        saveflag=0;
        return;
      }
    
    Filename=(char *)fl_get_input(ui->filename);
    if((fd=fopen(Filename,"r"))==(FILE *)NULL)
      {
        fl_show_object( ui->invalidgp);
        fprintf(stderr,"Unknown file %s\n",Filename);
        return;
      }
     
    if( fgets(dummy,80,fd)==(char *)NULL)
      { 
        fl_show_object(ui->invalidgp);
        fprintf(stderr,"File %s empty",Filename);
        fclose(fd);
        return;
      }
    for(i=1;i<31;i++)
      {
        float d0 ,d1;
        if (fgets(dummy,80,fd)==(char *)NULL)
          {
            fl_show_object(ui->invalidgp);
	    fclose (fd);
 	    return;
          }
     
        sscanf(dummy,"%g %g",&d0,&d1);
        ampl[0][i]=d0;
        ampl[1][i]=d1;
        /*
        fprintf(stderr,"%s",dummy);
        fprintf(stderr,"ampl[0][i]=%g ampl[1][i]=%g\n",(double)ampl[0][i],(double)ampl[1][i]);
        */
        fl_set_slider_value(Amp[i],ampl[0][i]);
        fl_set_slider_value(P[i],ampl[1][i]);
   
      }
    saveflag=loadflag=0;
    fclose(fd);
    fl_hide_object(ui->inputgroup);
    display();
}

/**************************************************************************
       Function save -- save spectrum to file
************************************************************************/      
void save (FL_OBJECT * savebut, long val)
{
   FILE * fd;
   int i;
   char dummy[80];
   if(saveflag==0)
     {   
       fl_show_object(ui->inputgroup);
       loadflag=0;
       saveflag=1;
       return;
     }
    
   Filename=(char *)fl_get_input(ui->filename);
   if((fd=fopen(Filename,"w+"))==(FILE *)NULL)
     {
        fl_show_object( ui->invalidgp);
        fprintf(stderr,"Cannot open or create file %s\n",Filename);
        return;
     }
    
     
   if( fprintf(fd,"Amplitude and Phase for synth\n")<=0)
     {
        fl_show_object(ui->invalidgp);
        fclose (fd);
        return;
     }
   for(i=1;i<31;i++)
     {
       if (fprintf(fd,"%g %g\n",(double)ampl[0][i],(double)ampl[1][i])<=0)
         {
           fl_show_object(ui->invalidgp);
	   fclose(fd);
	   return;
         }
   
      }
    fclose( fd);
    fl_hide_object(ui->inputgroup);
    loadflag=saveflag=0;
}

/**************************************************************************
       Function invalid -- hide  "Invalid File" message on button push
************************************************************************/
void invalid(FL_OBJECT * invalidbut, long val)
{
    saveflag=loadflag=0;
    fl_hide_object(ui->inputgroup);
    fl_hide_object(ui->invalidgp);
}

/***************************************************************************
        Function clip-- show clip level on display of graph.
***************************************************************************/
clip()
{
float vol,pampl,x[2],y[2];
vol=fl_get_slider_value(ui->volume); /*Volume of tone*/
     if( fl_get_button(ui->ampbut)==1) /* also amplify by display scale ? */
        pampl=exp(fl_get_slider_value(ui->scale)*log(2));
     else 
        pampl=1.;
  x[0]=40.;
  x[1]=0.;
  y[0]=y[1]=2.*pampl/vol;
    fl_add_xyplot_overlay(ui->xyplotgr,2,x,y,2,FL_RED);
    y[0]*=-1.;y[1]*=-1.;

    fl_add_xyplot_overlay(ui->xyplotgr,3,x,y,2,FL_RED);
}
/*******************************************************
   volume
   ********************************************************/
void volume_slide( FL_OBJECT * volume, long val)
  {
    clip();
  }  

/*****************************************************
   grab_warn_group
******************************************************/
/****************************************************
  Callback grab_ok_but
****************************************************/
void grab_ok_but( FL_OBJECT * dummy, long val)
{
  grab_ok_flag=1;
  fl_hide_object(ui->grab_warn_gp);
  grab(ui->grabbut,0);
}

/*********************************************************
  Callback grab_cancel_but
  *********************************************************/
void grab_cancel_but(FL_OBJECT * dummy,long val)
{
  grab_ok_flag=0;
  fl_hide_object(ui->grab_warn_gp);

}

/**********************************************************
   Function spectrum-- take NMAX point spectrum of recorded sound
     Use the full screen display.(Note aliasing severe if frequency of
     componenets exceeds max of spectrum)--square window function.
   ********************************************************/
spectrum()
{
      float ys[NMAX],xs[NMAX],j,j0,dj;
      long i;
/* Used for testing*//*
      for(i=0;i<xmax[level];i++)
        yrd[i]= sin( (double)i*2.*PI/RATE*RATE/44.1);
 */    
      j=j0=xmin[level];
      dj=(xmax[level]-xmin[level])/(float)(NMAX-1);

      if (fabs( (double)(xmax[level]-xmin[level]))>NMAX)/*  disable*/
        { 
         /*Take average of the points to right of the point in question*/ 	
	   int ii;
           float drj;
           dj=(xmax[level]-xmin[level])/(float)(NMAX+.01);
          for (i=0;i<NMAX ;i++)
            {
               drj=j-floor(j);
               ys[i]=((1.-drj)*(1.-drj)*yrd[(int)j]-drj*drj*yrd[(int)j+1])/(2.*dj);
               for (ii=1;(int)j+ii<=(int)(j+dj);ii++)
                 ys[i]+=yrd[(int)j+ii]/(dj);          
               j+=dj;
               if((drj=j-floor(j))>0)
                  ys[i]+= (-(1.-drj)*(1-drj)*yrd[(int)j] +drj*drj*yrd[(int)j+1])/(2*dj);
               dum[i]=0.;
            }
        }
      else
	{

          for(i=0;i<NMAX;i++)
            {
	       ys[i]= yrd[(int)j]*( 1.-(j-floor(j)))+yrd[(int)j+1]*(j-floor(j));
	       j+=dj;
               dum[i]=0;

            }

        }
	
	

      fft(ys,dum,four[0],four[1],NMAX,1);
      for (i=0;i<NMAX; i++)
        {
          ys[i]=log10(2.*(four[0][i]*four[0][i]+four[1][i]*four[1][i]))*10.;
          xs[i]=(float)i*RATE/(dj*NMAX);
	  if(ys[i]<-60) ys[i]=-60;
        }
 

      fl_set_xyplot_xscale(ui->fourgraph, ll_flag, 10.);
      
      fl_set_xyplot_xtics(ui->fourgraph,5,5);
      fl_set_xyplot_ybounds(ui->fourgraph,-60,0);
      
      fl_set_xyplot_ytics(ui->fourgraph,6,5);
      

      fl_set_xyplot_data(ui->fourgraph,xs+1,ys+1,NMAX/2,"","Hz","dB");
      fl_show_object(ui->ll_button);

}
/**********************************************************************
   Callback fourier_but
***********************************************************************/
void fourier_but(FL_OBJECT * dummy,long val)
{
    if(fourier_flag==0&&recordstate>0)
      {
        fl_hide_object(ui->slider_group);
        fl_show_object(ui->fourier); 
        spectrum();

	fourier_flag=1;
	fl_set_object_label(ui->fourier_button,"Cancel\n FFT");

      }
    else
      {
        fl_hide_object(ui->fourier);
	fourier_flag=0;
	fl_set_object_label(ui->fourier_button,"FFT");
        fl_set_positioner_xvalue(ui->spectrum_posn,0.); 
	fl_show_object(ui->slider_group); 
      }

}

/********************************************************************
Callback function linear_log_but
*********************************************************************/	   
void linear_log_but(FL_OBJECT * dummy,long val)
{
     if(ll_flag==FL_LINEAR)
       { ll_flag=FL_LOG;
         fl_set_object_label(ui->ll_button,"Linear");
       }
     else
       { ll_flag=FL_LINEAR;
         fl_set_object_label(ui->ll_button,"Log");
       }
     fl_set_positioner_xvalue(ui->spectrum_posn,0.);  
     spectrum();
}

/******************************************************************
Callback spectrum_position
*******************************************************************/

void spectrum_position(FL_OBJECT * Pos,long val)
{
    float X,Y,a,b;
    FL_Coord px,py;
    unsigned int k;
    char freq[80];
    fl_get_win_mouse(synth_window,&px,&py,&k);
    fl_xyplot_s2w(ui->fourgraph, (double )px, (double) py,&X,&Y);
    
    sprintf(freq,"%.1f Hz",(double)X);
    fl_set_object_label(ui->frequency,freq);
    fl_show_object(ui->frequency);
    
}
